module RBBST (Tree, empty, isEmpty, insert, merge, member) where

import Test.QuickCheck
import Control.Monad

data Tree a = Node (Tree a) (a,Colour) (Tree a) | Empty
            deriving (Show, Eq)

data Colour = Red | Black
            deriving (Show, Eq)

empty :: Tree a
empty = Empty

isEmpty :: Tree a -> Bool
isEmpty Empty = True
isEmpty _     = False

colour (Node _ (_,c) _) = c

insert x t = makeBlack $ insert' x t

insert' :: Ord a => a -> Tree a -> Tree a
insert' x Empty = Node Empty (x,Red) Empty
insert' x t@(Node lt (y,Black) rt)
  | x == y    = t
  | x <  y    = balance (insert' x lt) y rt
  | otherwise = balance lt y (insert' x rt)
insert' x t@(Node lt (y,Red) rt)
  | x == y    = t
  | x <  y    = Node (insert' x lt) (y,Black) rt
  | otherwise = Node lt (y,Red) (insert' x rt)

memberC :: Ord a => a -> Maybe a -> Tree a -> Bool
memberC x Nothing  Empty          = False
memberC x (Just c) Empty          = x == c
memberC x c (Node lt (y,_) rt)
  | x < y = memberC x c lt
  | otherwise = memberC x (Just y) rt

member :: Ord a => a -> Tree a -> Bool
member x = memberC x Nothing

flatten :: Tree a -> [a]
flatten Empty = []
flatten (Node lt (x,_) rt) = flatten lt ++ x : flatten rt

merge :: Ord a => Tree a -> Tree a -> Tree a
merge t1 t2 = foldr insert t1 (tail (flatten t2))

-- Red-black balancing
makeBlack (Node lt (y,_) rt) = Node lt (y,Red) rt

balance (Node (Node a (x,Red) b) (y,Red) c) z d =
  Node (Node a (x,Black) b) (y,Red) (Node c (z,Black) d)
balance (Node a (x,Red) (Node b (y,Red) c)) z d =
  Node (Node a (x,Black) b) (y,Red) (Node c (z,Black) d)
balance a x (Node (Node b (y,Red) c) (z,Red) d) =
  Node (Node a (x,Black) b) (y,Red) (Node c (z,Black) d)
balance a x (Node b (y,Red) (Node c (z,Red) d)) =
  Node (Node a (x,Black) b) (y,Red) (Node c (z,Black) d)
balance a x b = Node a (x,Black) b

-- Write the properties required in the assignment description